/* Copyright (c) Determina Corp. 2006.  All rights reserved.
 * 
 * This software and documentation constitute an unpublished work and contain
 * valuable trade secrets and proprietary information belonging to Determina
 * Corp.  None of the foregoing material may be copied, duplicated or
 * disclosed without the express written permission of Determina Corp.
 * DETERMINA CORP. EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING THIS
 * SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF MERCHANTABILITY
 * AND/OR FITNESS FOR ANY PARTICULAR PURPOSE, AND WARRANTIES OF PERFORMANCE,
 * AND ANY WARRANTY THAT MIGHT OTHERWISE ARISE FROM COURSE OF DEALING OR USAGE
 * OF TRADE. NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH RESPECT TO THE USE
 * OF THE SOFTWARE OR DOCUMENTATION. Under no circumstances shall Determina
 * Corp. be liable for incidental, special, indirect, direct or consequential
 * damages or loss of profits, interruption of business, or related expenses
 * which may arise from use of software or documentation, including but not
 * limited to those resulting from defects in software and/or documentation,
 * or loss or inaccuracy of data of any kind.
 */


#include <windows.h>
#include <winnt.h>

//
// Structures for module matching
//

struct hash {
    unsigned int start;
    unsigned int length;
    unsigned int crc32;
};

struct module {
    char* name;                 // DLL name
    struct hash hash;           // CRC32 checksum of the area around the patch
    unsigned int patch_point;   // offset of the patch point
    unsigned int jump_target;   // offset of the jump target
};


//
// Modules that we watch to patch
//

struct module modules[] = {

    // Windows 2000 SP0

    {
        "MSHTML.dll",
        { 0x1BC38B, 0x67, 3703781640 },
        0x1bc3d2,
        0x1bc3ab,
    },

    // Windows 2000 SP1

    {
        "MSHTML.dll",
        { 0x1AE6F3, 0x67, 239260724 },
        0x1ae73a,
        0x1ae713,
    },

    // MS01-015

    {
        "MSHTML.dll",
        { 0x1B6D84, 0x67, 3507111965 },
        0x1b6dcb,
        0x1b6da4,
    },

    // Windows 2000 SP2

    {
        "MSHTML.dll",
        { 0x1AC61A, 0x67, 513595816 },
        0x1ac661,
        0x1ac63a,
    },

    // Windows 2000 SP2 SRP1

    {
        "MSHTML.dll",
        { 0x1AB935, 0x67, 1859291035 },
        0x1ab97c,
        0x1ab955,
    },

    // MS02-005

    {
        "MSHTML.dll",
        { 0x176E7C, 0x4E, 119161652 },
        0x176eaa,
        0x176e9c,
    },

    // Windows 2000 SP3

    {
        "MSHTML.dll",
        { 0x1A8ADB, 0x67, 2197965368 },
        0x1a8b22,
        0x1a8afb,
    },

    // MS02-047

    {
        "MSHTML.dll",
        { 0x1B46D0, 0x67, 3487582208 },
        0x1b4717,
        0x1b46f0,
    },

    // Internet Explorer 5.01 SP2-SP3 MS04-025

    {
        "MSHTML.dll",
        { 0x177B6A, 0x4E, 2207989960 },
        0x177b98,
        0x177b8a,
    },

    // Internet Explorer 5.01 SP3 MS04-038

    {
        "MSHTML.dll",
        { 0x17828E, 0x4E, 82036462 },
        0x1782bc,
        0x1782ae,
    },

    // Internet Explorer 5.01 SP3 MS05-014

    {
        "MSHTML.dll",
        { 0x1787BC, 0x4E, 855560832 },
        0x1787ea,
        0x1787dc,
    },

    // Internet Explorer 5.01 SP3 MS05-020

    {
        "MSHTML.dll",
        { 0x17932C, 0x4E, 1826811257 },
        0x17935a,
        0x17934c,
    },

    // Internet Explorer 5.01 SP3 MS05-025

    {
        "MSHTML.dll",
        { 0x17947F, 0x4E, 374580404 },
        0x1794ad,
        0x17949f,
    },

    // Windows 2000 SP4

    {
        "MSHTML.dll",
        { 0x1AE5C3, 0x67, 2678223039 },
        0x1ae60a,
        0x1ae5e3,
    },

    // Internet Explorer 5.01 SP4 MS04-025

    {
        "MSHTML.dll",
        { 0x177B4B, 0x4E, 976821450 },
        0x177b79,
        0x177b6b,
    },

    // Internet Explorer 5.01 SP4 MS04-038

    {
        "MSHTML.dll",
        { 0x17828E, 0x4E, 82036462 },
        0x1782bc,
        0x1782ae,
    },

    // Internet Explorer 5.01 SP4 MS05-014

    {
        "MSHTML.dll",
        { 0x1787BC, 0x4E, 855560832 },
        0x1787ea,
        0x1787dc,
    },

    // Internet Explorer 5.01 SP4 MS05-020

    {
        "MSHTML.dll",
        { 0x17932C, 0x4E, 1826811257 },
        0x17935a,
        0x17934c,
    },

    // Windows 2000 SP4 UR1

    {
        "MSHTML.dll",
        { 0x17932C, 0x4E, 1826811257 },
        0x17935a,
        0x17934c,
    },

    // Internet Explorer 5.01 SP4 MS05-025

    {
        "MSHTML.dll",
        { 0x17947F, 0x4E, 374580404 },
        0x1794ad,
        0x17949f,
    },

    // Internet Explorer 5.01 SP4 MS05-038

    {
        "MSHTML.dll",
        { 0x17950D, 0x4E, 1849753469 },
        0x17953b,
        0x17952d,
    },

    // Internet Explorer 5.01 SP4 MS05-052

    {
        "MSHTML.dll",
        { 0x1794B6, 0x4E, 3042377783 },
        0x1794e4,
        0x1794d6,
    },

    // Internet Explorer 5.01 SP4 MS05-054

    {
        "MSHTML.dll",
        { 0x179459, 0x4E, 2059396922 },
        0x179487,
        0x179479,
    },

    // Internet Explorer 5.01 SP4 MS06-004

    {
        "MSHTML.dll",
        { 0x179459, 0x4E, 2059396922 },
        0x179487,
        0x179479,
    },

    // Windows XP SP0

    {
        "MSHTML.dll",
        { 0x23F59D, 0x4E, 1930178703 },
        0x23f5cb,
        0x23f5bd,
    },

    // MS02-027

    {
        "MSHTML.dll",
        { 0x233A90, 0x4E, 463211394 },
        0x233abe,
        0x233ab0,
    },

    // Internet Explorer 6 SP0 MS04-025

    {
        "MSHTML.dll",
        { 0x237DC6, 0x4E, 3435046334 },
        0x237df4,
        0x237de6,
    },

    // Internet Explorer 6 SP0 MS04-038

    {
        "MSHTML.dll",
        { 0x236A9B, 0x4E, 1020853566 },
        0x236ac9,
        0x236abb,
    },

    // Windows XP SP1a

    {
        "MSHTML.dll",
        { 0x241256, 0x4E, 4224144104 },
        0x241284,
        0x241276,
    },

    // MS02-068

    {
        "MSHTML.dll",
        { 0x236B28, 0x4E, 872074275 },
        0x236b56,
        0x236b48,
    },

    // MS03-004

    {
        "MSHTML.dll",
        { 0x231C00, 0x4E, 490751674 },
        0x231c2e,
        0x231c20,
    },

    // MS03-015, MS03-020

    {
        "MSHTML.dll",
        { 0x233B9B, 0x4E, 2130636563 },
        0x233bc9,
        0x233bbb,
    },

    // MS03-048

    {
        "MSHTML.dll",
        { 0x239E13, 0x4E, 2689805499 },
        0x239e41,
        0x239e33,
    },

    // MS04-004

    {
        "MSHTML.dll",
        { 0x2380A8, 0x4E, 4014561778 },
        0x2380d6,
        0x2380c8,
    },

    // Internet Explorer 6 SP1 MS04-025

    {
        "MSHTML.dll",
        { 0x23AE2E, 0x4E, 970562258 },
        0x23ae5c,
        0x23ae4e,
    },

    // Internet Explorer 6 SP1 MS04-038 Windows-2000-XP, MS04-038 Windows-NT4sp6a-98-ME

    {
        "MSHTML.dll",
        { 0x23B636, 0x4E, 1263864362 },
        0x23b664,
        0x23b656,
    },

    // Internet Explorer 6 SP1 MS04-040 Windows-2000-XP, MS04-040 Windows-NT4sp6a-98-ME

    {
        "MSHTML.dll",
        { 0x1DEF33, 0x4E, 1751273106 },
        0x1def61,
        0x1def53,
    },

    // Internet Explorer 6 SP1 MS05-014 gdr Windows-2000-XP

    {
        "MSHTML.dll",
        { 0x23A98B, 0x4E, 2165099349 },
        0x23a9b9,
        0x23a9ab,
    },

    // Internet Explorer 6 SP1 MS05-014 qfe Windows-2000-XP

    {
        "MSHTML.dll",
        { 0x1E01C6, 0x4E, 2312365653 },
        0x1e01f4,
        0x1e01e6,
    },

    // Internet Explorer 6 SP1 MS05-020 gdr Windows-2000-XP

    {
        "MSHTML.dll",
        { 0x23D3EE, 0x4E, 40563454 },
        0x23d41c,
        0x23d40e,
    },

    // Internet Explorer 6 SP1 MS05-020 qfe Windows-2000-XP

    {
        "MSHTML.dll",
        { 0x1E0CB8, 0x4E, 1743158687 },
        0x1e0ce6,
        0x1e0cd8,
    },

    // Internet Explorer 6 SP1 MS05-025 gdr Windows-2000-XP

    {
        "MSHTML.dll",
        { 0x1E00D5, 0x4E, 3567684944 },
        0x1e0103,
        0x1e00f5,
    },

    // Internet Explorer 6 SP1 MS05-025 qfe Windows-2000-XP

    {
        "MSHTML.dll",
        { 0x1E10E4, 0x4E, 633659183 },
        0x1e1112,
        0x1e1104,
    },

    // Internet Explorer 6 SP1 MS05-038 gdr Windows-2000-XP

    {
        "MSHTML.dll",
        { 0x1E0033, 0x4E, 980492186 },
        0x1e0061,
        0x1e0053,
    },

    // Internet Explorer 6 SP1 MS05-038 qfe Windows-2000-XP

    {
        "MSHTML.dll",
        { 0x1E12E0, 0x4E, 2911226202 },
        0x1e130e,
        0x1e1300,
    },

    // Internet Explorer 6 SP1 MS05-052 gdr Windows-2000-XP

    {
        "MSHTML.dll",
        { 0x1E01C3, 0x4E, 4229070414 },
        0x1e01f1,
        0x1e01e3,
    },

    // Internet Explorer 6 SP1 MS05-052 qfe Windows-2000-XP

    {
        "MSHTML.dll",
        { 0x1E12C9, 0x4E, 325362658 },
        0x1e12f7,
        0x1e12e9,
    },

    // Internet Explorer 6 SP1 MS05-054 gdr

    {
        "MSHTML.dll",
        { 0x1E01F4, 0x4E, 4052500837 },
        0x1e0222,
        0x1e0214,
    },

    // Internet Explorer 6 SP1 MS05-054 qfe

    {
        "MSHTML.dll",
        { 0x1E13BF, 0x4E, 1170668494 },
        0x1e13ed,
        0x1e13df,
    },

    // Windows XP SP2

    {
        "MSHTML.dll",
        { 0x23C257, 0x4D, 2200774393 },
        0x23c284,
        0x23c277,
    },

    // Windows XP SP2 MS04-038 gdr

    {
        "MSHTML.dll",
        { 0x23C3D7, 0x4D, 3523596815 },
        0x23c404,
        0x23c3f7,
    },

    // Windows XP SP2 MS04-038 qfe

    {
        "MSHTML.dll",
        { 0x23C3D7, 0x4D, 3523596815 },
        0x23c404,
        0x23c3f7,
    },

    // Windows XP SP2 MS05-014 gdr

    {
        "MSHTML.dll",
        { 0x23CDE6, 0x4D, 1077742785 },
        0x23ce13,
        0x23ce06,
    },

    // Windows XP SP2 MS05-014 qfe

    {
        "MSHTML.dll",
        { 0x23D147, 0x4D, 1717860001 },
        0x23d174,
        0x23d167,
    },

    // Windows XP SP2 MS05-020 gdr

    {
        "MSHTML.dll",
        { 0x23DC8F, 0x4D, 1508777697 },
        0x23dcbc,
        0x23dcaf,
    },

    // Windows XP SP2 MS05-020 qfe

    {
        "MSHTML.dll",
        { 0x23DDFF, 0x4D, 2369711915 },
        0x23de2c,
        0x23de1f,
    },

    // Windows XP SP2 MS05-025 gdr

    {
        "MSHTML.dll",
        { 0x23DCDE, 0x4D, 2596397986 },
        0x23dd0b,
        0x23dcfe,
    },

    // Windows XP SP2 MS05-025 qfe

    {
        "MSHTML.dll",
        { 0x23E227, 0x4D, 3594117519 },
        0x23e254,
        0x23e247,
    },

    // Windows XP SP2 MS05-038 gdr

    {
        "MSHTML.dll",
        { 0x23DEFC, 0x4D, 2662725898 },
        0x23df29,
        0x23df1c,
    },

    // Windows XP SP2 MS05-038 qfe

    {
        "MSHTML.dll",
        { 0x23E5DE, 0x4D, 3684116905 },
        0x23e60b,
        0x23e5fe,
    },

    // Windows XP SP2 MS05-052 gdr

    {
        "MSHTML.dll",
        { 0x23E006, 0x4D, 4240162876 },
        0x23e033,
        0x23e026,
    },

    // Windows XP SP2 MS05-052 qfe

    {
        "MSHTML.dll",
        { 0x23E73F, 0x4D, 349972183 },
        0x23e76c,
        0x23e75f,
    },

    // Windows XP SP2 MS05-054 gdr

    {
        "MSHTML.dll",
        { 0x23E0E6, 0x4D, 3369728779 },
        0x23e113,
        0x23e106,
    },

    // Windows XP SP2 MS05-054 qfe

    {
        "MSHTML.dll",
        { 0x23E947, 0x4D, 3316967350 },
        0x23e974,
        0x23e967,
    },

    // Windows 2003 SP0

    {
        "MSHTML.dll",
        { 0x21F51A, 0x4E, 3101617912 },
        0x21f548,
        0x21f53a,
    },

    // Windows 2003 SP0 MS03-032 gdr

    {
        "MSHTML.dll",
        { 0x248446, 0x4E, 2687869807 },
        0x248474,
        0x248466,
    },

    // Windows 2003 SP0 MS03-032 qfe

    {
        "MSHTML.dll",
        { 0x24845E, 0x4E, 365648356 },
        0x24848c,
        0x24847e,
    },

    // Windows 2003 SP0 MS03-040 gdr

    {
        "MSHTML.dll",
        { 0x2485FE, 0x4E, 611323582 },
        0x24862c,
        0x24861e,
    },

    // Windows 2003 SP0 MS03-040 qfe

    {
        "MSHTML.dll",
        { 0x248616, 0x4E, 2540528212 },
        0x248644,
        0x248636,
    },

    // Windows 2003 SP0 MS03-048 gdr

    {
        "MSHTML.dll",
        { 0x248762, 0x4E, 2758358678 },
        0x248790,
        0x248782,
    },

    // Windows 2003 SP0 MS03-048 qfe

    {
        "MSHTML.dll",
        { 0x24877A, 0x4E, 3007013436 },
        0x2487a8,
        0x24879a,
    },

    // Windows 2003 SP0 MS04-004 gdr

    {
        "MSHTML.dll",
        { 0x248866, 0x4E, 1885868963 },
        0x248894,
        0x248886,
    },

    // Windows 2003 SP0 MS04-004 qfe

    {
        "MSHTML.dll",
        { 0x2488E6, 0x4E, 2294026423 },
        0x248914,
        0x248906,
    },

    // Windows 2003 SP0 MS04-025 gdr

    {
        "MSHTML.dll",
        { 0x248C62, 0x4E, 3869614450 },
        0x248c90,
        0x248c82,
    },

    // Windows 2003 SP0 MS04-025 qfe

    {
        "MSHTML.dll",
        { 0x248D02, 0x4E, 2700948646 },
        0x248d30,
        0x248d22,
    },

    // Windows 2003 SP0 MS04-038 gdr

    {
        "MSHTML.dll",
        { 0x249432, 0x4E, 1960412349 },
        0x249460,
        0x249452,
    },

    // Windows 2003 SP0 MS04-038 qfe

    {
        "MSHTML.dll",
        { 0x2495A6, 0x4E, 1931202151 },
        0x2495d4,
        0x2495c6,
    },

    // Windows 2003 SP0 MS05-014 gdr

    {
        "MSHTML.dll",
        { 0x249A36, 0x4E, 1113017411 },
        0x249a64,
        0x249a56,
    },

    // Windows 2003 SP0 MS05-014 qfe

    {
        "MSHTML.dll",
        { 0x249CB2, 0x4E, 39789870 },
        0x249ce0,
        0x249cd2,
    },

    // Windows 2003 SP0 MS05-020 gdr

    {
        "MSHTML.dll",
        { 0x24A642, 0x4E, 1216577486 },
        0x24a670,
        0x24a662,
    },

    // Windows 2003 SP0 MS05-020 qfe

    {
        "MSHTML.dll",
        { 0x24A7A2, 0x4E, 3780654531 },
        0x24a7d0,
        0x24a7c2,
    },

    // Windows 2003 SP0 MS05-025 gdr

    {
        "MSHTML.dll",
        { 0x24A7AA, 0x4E, 1043652095 },
        0x24a7d8,
        0x24a7ca,
    },

    // Windows 2003 SP0 MS05-025 qfe

    {
        "MSHTML.dll",
        { 0x24AB4E, 0x4E, 951942412 },
        0x24ab7c,
        0x24ab6e,
    },

    // Windows 2003 SP0 MS05-038 gdr

    {
        "MSHTML.dll",
        { 0x24A862, 0x4E, 1875965646 },
        0x24a890,
        0x24a882,
    },

    // Windows 2003 SP0 MS05-038 qfe

    {
        "MSHTML.dll",
        { 0x24AC9E, 0x4E, 3589421852 },
        0x24accc,
        0x24acbe,
    },

    // Windows 2003 SP0 MS05-052 gdr

    {
        "MSHTML.dll",
        { 0x24A9EE, 0x4E, 597879652 },
        0x24aa1c,
        0x24aa0e,
    },

    // Windows 2003 SP0 MS05-052 qfe

    {
        "MSHTML.dll",
        { 0x24AE46, 0x4E, 2230516184 },
        0x24ae74,
        0x24ae66,
    },

    // Windows 2003 SP0 MS05-054 gdr

    {
        "MSHTML.dll",
        { 0x22125E, 0x4E, 1174339943 },
        0x22128c,
        0x22127e,
    },

    // Windows 2003 SP0 MS05-054 qfe

    {
        "MSHTML.dll",
        { 0x221676, 0x4E, 2911951141 },
        0x2216a4,
        0x221696,
    },

    // Windows 2003 SP1

    {
        "MSHTML.dll",
        { 0x2660BD, 0x57, 2186665457 },
        0x2660f4,
        0x2660dd,
    },

    // Windows 2003 SP1 MS05-025 gdr

    {
        "MSHTML.dll",
        { 0x2662A2, 0x57, 2922979747 },
        0x2662d9,
        0x2662c2,
    },

    // Windows 2003 SP1 MS05-025 qfe

    {
        "MSHTML.dll",
        { 0x2662A2, 0x57, 2922979747 },
        0x2662d9,
        0x2662c2,
    },

    // Windows 2003 SP1 MS05-038 gdr

    {
        "MSHTML.dll",
        { 0x26651E, 0x57, 4220975131 },
        0x266555,
        0x26653e,
    },

    // Windows 2003 SP1 MS05-038 qfe

    {
        "MSHTML.dll",
        { 0x2665F9, 0x57, 4198091938 },
        0x266630,
        0x266619,
    },

    // Windows 2003 SP1 MS05-052 gdr

    {
        "MSHTML.dll",
        { 0x266B85, 0x57, 1696029311 },
        0x266bbc,
        0x266ba5,
    },

    // Windows 2003 SP1 MS05-052 qfe

    {
        "MSHTML.dll",
        { 0x266CB6, 0x57, 2111726732 },
        0x266ced,
        0x266cd6,
    },

    // Windows 2003 SP1 MS05-054 gdr

    {
        "MSHTML.dll",
        { 0x266BC8, 0x57, 675819495 },
        0x266bff,
        0x266be8,
    },

    // Windows 2003 SP1 MS05-054 qfe

    {
        "MSHTML.dll",
        { 0x266DEC, 0x57, 2134024224 },
        0x266e23,
        0x266e0c,
    }
};

#define NUM_MODULES (sizeof(modules)/sizeof(struct module))


/* We want to patch the CInput::createTextRange function in mshtml.dll. The
   code that we are interested in looks like this:

.text:7D6DC270 33 FF                          xor     edi, edi
.text:7D6DC272 39 75 0C                       cmp     [ebp+arg_4], esi
.text:7D6DC275 75 0A                          jnz     short arg_4_not_NULL
.text:7D6DC277 BF 57 00 07 80                 mov     edi, 80070057h
.text:7D6DC27C E9 88 00 00 00                 jmp     set_error_and_return
.text:7D6DC281
.text:7D6DC281                arg_4_not_NULL:
.text:7D6DC281 8B 4D 08                       mov     ecx, [ebp+arg_0]
.text:7D6DC284 F6 41 18 20                    test    byte ptr [ecx+18h], 20h
.text:7D6DC288 74 7F                          jz      short set_error_and_return

        ...                                   ...

.text:7D6DC309                set_error_and_return:
.text:7D6DC309 8B 4D 08                       mov     ecx, [ebp+arg_0]
.text:7D6DC30C 57                             push    edi
.text:7D6DC30D E8 49 20 E2 FF                 call    CBase::SetErrorInfo
.text:7D6DC312 5F                             pop     edi
.text:7D6DC313 5E                             pop     esi
.text:7D6DC314 5D                             pop     ebp
.text:7D6DC315 C2 08 00                       retn    8

   The code above performs two checks on the function arguments. If arg_4 is
   NULL, the error code is set to 0x80070057 (E_INVALIDARG in WinError.h) and
   the function returns. If one of the bits in the arg_0 structure is not set,
   the function also returns, but the error code is not specified. Since edi
   was previously initialzed to 0, the function returns 0, indicating success.
   Returning success instead of an error leads to memory corruption later.

   We will patch the jz instruction after the test and change its target to the
   mov edi, 80070057h instruction. In most cases this can be done by patching a
   single byte in the executable.

   One complication is that in some versions of mshtml.dll the jz instruction
   is a short jump with an 8-bit operand, and in other versions it's a near
   jump with a 32-bit operand. Our patching code handles both cases. */


// The test instruction is different for IE5 and IE6.

#define IE5_TEST "\xF6\x41\x19\x20"    // test byte ptr [ecx+19h], 20h
#define IE6_TEST "\xF6\x41\x18\x20"    // test byte ptr [ecx+18h], 20h

// Opcodes for the two forms of the JZ instruction

#define JZ_REL8   "\x74"                // jz short
#define JZ_REL32  "\x0F\x84"            // jz near


// Patch status

enum patch_status {
    NOT_PATCHED,
    PATCHED
};

extern "C" {
    __declspec(dllexport) enum patch_status patch_status = NOT_PATCHED;
}

//
// Patch the module. Returns FALSE if the patch could not be applied.
//

BOOL patch_module(struct module* module, void* image_base)
{
    // Get the absolute addresses of the patch point and jump target

    char* patch_point = (char*)image_base + module->patch_point;
    char* jump_target = (char*)image_base + module->jump_target;

    // Make sure that the patch point starts with a test instruction

    if (memcmp(patch_point, IE6_TEST, sizeof(IE6_TEST)-1) != 0 &&
        memcmp(patch_point, IE5_TEST, sizeof(IE5_TEST)-1) != 0)
    {
        return FALSE;
    }

    // Calculate the address of the jz operand and its length
    
    void* operand_addr;
    int operand_length;

    if (memcmp(patch_point+4, JZ_REL8, sizeof(JZ_REL8)-1) == 0) {
        operand_addr = patch_point + 4 + sizeof(JZ_REL8)-1;
        operand_length = 1;
    }
    else if (memcmp(patch_point+4, JZ_REL32, sizeof(JZ_REL32)-1) == 0) {
        operand_addr = patch_point + 4 + sizeof(JZ_REL32)-1;
        operand_length = 4;
    }
    else {
        // this is not a jz instruction
        return FALSE;
    }

    // Calculate the relative address of the jump target

    int jump_target_rel = (int)jump_target - ((int)operand_addr+operand_length);

    // If we're patching a short jump, the jump target must be within 128 bytes

    if (operand_length == 1 && (jump_target_rel > 0x7f || jump_target_rel < -0x80)) {
        return FALSE;
    }

    // Change the memory protection of the patch point to enable writing
    
    DWORD old_protection, dummy;

    if (VirtualProtect(operand_addr, operand_length, PAGE_EXECUTE_READWRITE, &old_protection) == 0) {
        return FALSE;
    }

    // Replace the target of the jz instruction with out target. This works
    // because we're on a little endian machine

    memcpy(operand_addr, &jump_target_rel, operand_length);

    // Restore the memory protection of the patch point

    VirtualProtect(operand_addr, operand_length, old_protection, &dummy);

    patch_status = PATCHED;

    return TRUE;
}


//
// Table for CRC32 calculation
//

const unsigned int crctab[] = {
	0x00000000, 0x77073096, 0xee0e612c, 0x990951ba,
	0x076dc419, 0x706af48f, 0xe963a535, 0x9e6495a3,
	0x0edb8832, 0x79dcb8a4, 0xe0d5e91e, 0x97d2d988,
	0x09b64c2b, 0x7eb17cbd, 0xe7b82d07, 0x90bf1d91,
	0x1db71064, 0x6ab020f2, 0xf3b97148, 0x84be41de,
	0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7,
	0x136c9856, 0x646ba8c0, 0xfd62f97a, 0x8a65c9ec,
	0x14015c4f, 0x63066cd9, 0xfa0f3d63, 0x8d080df5,
	0x3b6e20c8, 0x4c69105e, 0xd56041e4, 0xa2677172,
	0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b,
	0x35b5a8fa, 0x42b2986c, 0xdbbbc9d6, 0xacbcf940,
	0x32d86ce3, 0x45df5c75, 0xdcd60dcf, 0xabd13d59,
	0x26d930ac, 0x51de003a, 0xc8d75180, 0xbfd06116,
	0x21b4f4b5, 0x56b3c423, 0xcfba9599, 0xb8bda50f,
	0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924,
	0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d,
	0x76dc4190, 0x01db7106, 0x98d220bc, 0xefd5102a,
	0x71b18589, 0x06b6b51f, 0x9fbfe4a5, 0xe8b8d433,
	0x7807c9a2, 0x0f00f934, 0x9609a88e, 0xe10e9818,
	0x7f6a0dbb, 0x086d3d2d, 0x91646c97, 0xe6635c01,
	0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e,
	0x6c0695ed, 0x1b01a57b, 0x8208f4c1, 0xf50fc457,
	0x65b0d9c6, 0x12b7e950, 0x8bbeb8ea, 0xfcb9887c,
	0x62dd1ddf, 0x15da2d49, 0x8cd37cf3, 0xfbd44c65,
	0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2,
	0x4adfa541, 0x3dd895d7, 0xa4d1c46d, 0xd3d6f4fb,
	0x4369e96a, 0x346ed9fc, 0xad678846, 0xda60b8d0,
	0x44042d73, 0x33031de5, 0xaa0a4c5f, 0xdd0d7cc9,
	0x5005713c, 0x270241aa, 0xbe0b1010, 0xc90c2086,
	0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
	0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4,
	0x59b33d17, 0x2eb40d81, 0xb7bd5c3b, 0xc0ba6cad,
	0xedb88320, 0x9abfb3b6, 0x03b6e20c, 0x74b1d29a,
	0xead54739, 0x9dd277af, 0x04db2615, 0x73dc1683,
	0xe3630b12, 0x94643b84, 0x0d6d6a3e, 0x7a6a5aa8,
	0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1,
	0xf00f9344, 0x8708a3d2, 0x1e01f268, 0x6906c2fe,
	0xf762575d, 0x806567cb, 0x196c3671, 0x6e6b06e7,
	0xfed41b76, 0x89d32be0, 0x10da7a5a, 0x67dd4acc,
	0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5,
	0xd6d6a3e8, 0xa1d1937e, 0x38d8c2c4, 0x4fdff252,
	0xd1bb67f1, 0xa6bc5767, 0x3fb506dd, 0x48b2364b,
	0xd80d2bda, 0xaf0a1b4c, 0x36034af6, 0x41047a60,
	0xdf60efc3, 0xa867df55, 0x316e8eef, 0x4669be79,
	0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236,
	0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f,
	0xc5ba3bbe, 0xb2bd0b28, 0x2bb45a92, 0x5cb36a04,
	0xc2d7ffa7, 0xb5d0cf31, 0x2cd99e8b, 0x5bdeae1d,
	0x9b64c2b0, 0xec63f226, 0x756aa39c, 0x026d930a,
	0x9c0906a9, 0xeb0e363f, 0x72076785, 0x05005713,
	0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38,
	0x92d28e9b, 0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21,
	0x86d3d2d4, 0xf1d4e242, 0x68ddb3f8, 0x1fda836e,
	0x81be16cd, 0xf6b9265b, 0x6fb077e1, 0x18b74777,
	0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c,
	0x8f659eff, 0xf862ae69, 0x616bffd3, 0x166ccf45,
	0xa00ae278, 0xd70dd2ee, 0x4e048354, 0x3903b3c2,
	0xa7672661, 0xd06016f7, 0x4969474d, 0x3e6e77db,
	0xaed16a4a, 0xd9d65adc, 0x40df0b66, 0x37d83bf0,
	0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
	0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6,
	0xbad03605, 0xcdd70693, 0x54de5729, 0x23d967bf,
	0xb3667a2e, 0xc4614ab8, 0x5d681b02, 0x2a6f2b94,
	0xb40bbe37, 0xc30c8ea1, 0x5a05df1b, 0x2d02ef8d
};


//
// This function implements the Ethernet AUTODIN II CRC32 algorithm.
//

unsigned int crc32(const char* buf, const unsigned int len)
{
    unsigned int i;
    unsigned int crc = 0xFFFFFFFF;

    for (i = 0; i < len; i++)
        crc = (crc >> 8) ^ crctab[(crc ^ buf[i]) & 0xFF];

    return crc;
}


//
// Test if an address is inside a loaded PE image. Returns TRUE or FALSE
//

BOOL addr_in_image(void* addr, void* image_base, DWORD image_size)
{
    if (addr == NULL || image_base == NULL || image_size <= 0)
        return FALSE;

    if ((char*) addr < (char*) image_base ||
        (char*) addr > ((char*) image_base + image_size)) {
        return FALSE;
    }

    // FIXME: do a QueryVirtualMemory() & check for image; to be safe
    return TRUE;
}


//
// Get the DLL name of a loaded PE image. Returns NULL on error
//

char* get_dll_name(void* image_base, DWORD image_size, IMAGE_NT_HEADERS* nt)
{
    IMAGE_DATA_DIRECTORY *expdir = nt->OptionalHeader.DataDirectory +
                                   IMAGE_DIRECTORY_ENTRY_EXPORT;
    IMAGE_EXPORT_DIRECTORY *exports = NULL;
    char* dll_name = NULL;

    if (!addr_in_image((void*) expdir, image_base, image_size))
        return NULL;

    // We have a valid exports directory entry in data direcotry.
    if (expdir->Size > 0) {
        exports = (IMAGE_EXPORT_DIRECTORY *) ((char*) image_base +
                                              expdir->VirtualAddress);
    } else {
        return NULL;
    }

    if (!addr_in_image((void*) exports, image_base, image_size))
        return NULL;

    // We have a valid exports direcotry.
    dll_name =  (char*) image_base + exports->Name /* RVA */;
    if (!addr_in_image((void*) dll_name, image_base, image_size)) {
        return NULL;
    } else {
        return dll_name;
    }
}


//
// Matches the DLL loaded at the specified image base against the list of
// modules we want to patch. If the DLL matches, returns a pointer to the
// module structure, otherwise returns NULL.
//

struct module* module_match(void* image_base)
{
    IMAGE_DOS_HEADER* dos = (IMAGE_DOS_HEADER*) image_base;
    IMAGE_NT_HEADERS* nt;
    DWORD image_size;
    char* dll_name = NULL;
    unsigned int image_crc32 = 0;
    int i;

    // FIXME: assumes readability of header (dos, nt & optional) memory - 
    //      validate.

    // Is this an image of a PE file?
    if (image_base == NULL || dos->e_magic != IMAGE_DOS_SIGNATURE)
        return NULL;

    nt = (IMAGE_NT_HEADERS*) (((unsigned int)dos) + dos->e_lfanew);
    if (nt == NULL || nt->Signature != IMAGE_NT_SIGNATURE)
        return NULL;

    image_size = nt->OptionalHeader.SizeOfImage;
    if (image_size <= 0)
        return NULL;

    // Well, looks like an image, let us see if it has a PE name.
    dll_name = get_dll_name(image_base, image_size, nt);
    if (dll_name == NULL)
        return NULL;

    // Let us see if the image matches any of our DLLs.
    for (i = 0; i < NUM_MODULES; i++) {

        // It is ok to use strlen on the argument because it is a compile time
        // constant; the compiler should replace the call with a constant.
        if (memcmp((void*) dll_name, (void*) modules[i].name,
                   strlen(modules[i].name)) == 0) {
            char *patch_point = (char*)image_base + modules[i].patch_point;
            char *jump_target = (char*)image_base + modules[i].jump_target;
            char *hash_start = (char*)image_base + modules[i].hash.start;
            unsigned int hash_size = modules[i].hash.length;

            // Ok, dll name matched.  Now, let us verify hash bounds.
            if (!addr_in_image((void*) hash_start, image_base, image_size))
                continue;
            if (!addr_in_image((void*) (hash_start + hash_size),
                               image_base, image_size))
                continue;

            // Check the CRC32 hash
            image_crc32 = crc32(hash_start, hash_size);
            if (image_crc32 != modules[i].hash.crc32)
                continue;

            // Verify that the patch point and jump target are within bounds
            if (!addr_in_image((void*) (patch_point+10),
                               image_base, image_size))
                continue;

            if (!addr_in_image((void*) (jump_target),
                               image_base, image_size))
                continue;

            // We have a matching module
            return &modules[i];
        }
    }

    return NULL;
}


//
// Internal structures used for walking the loaded modules list
//

typedef struct {
    ULONG Length;
    BOOLEAN Initialized;
    PVOID SsHandle;
    LIST_ENTRY InLoadOrderModuleList;
    LIST_ENTRY InMemoryOrderModuleList;
    LIST_ENTRY InInitializationOrderModuleList;
} PEB_LDR_DATA;

typedef struct {
    LIST_ENTRY InLoadOrderModuleList;
    LIST_ENTRY InMemoryOrderModuleList;
    LIST_ENTRY InInitializationOrderModuleList;
    PVOID BaseAddress;
    // incomplete, but we don't need the rest
} LDR_MODULE;


//
// Walk the list of loaded DLLs and patch the ones that match
//

void patch_loaded_modules()
{
    PEB_LDR_DATA *ldr;
    LIST_ENTRY *first, *e;
    LDR_MODULE *mod;

    __asm {
        mov eax, fs:[0x30]  // get the PEB pointer
        mov eax, [eax+0xC]  // get the pointer to the loader data structure
        mov ldr, eax
    }

    first = &ldr->InLoadOrderModuleList;

    int count = 0;

    for (e = first->Flink; e != first; e = e->Flink)
    {
        mod = (LDR_MODULE*)e;

        // Patch the module if it matches

        struct module* module = module_match(mod->BaseAddress);

        if (module != NULL)
            patch_module(module, mod->BaseAddress);

        // Try to detect infinite loops
        
        if (count++ > 10000) {
            return;
        }
    }
}


//
// If our loader hook was successfully inserted, this function will be called
// every time the loader calls the DllMain routine of any loaded DLL. This
// allows us to intercept DLL loading and unloading. The arguments are the same
// as the ones passed to the real DllMain function.
//

void WINAPI DllMainHook(HINSTANCE hinstDll, DWORD fdwReason, LPVOID lpvReserved)
{
    UNREFERENCED_PARAMETER(fdwReason);
    UNREFERENCED_PARAMETER(lpvReserved);

    switch (fdwReason) {
        case DLL_PROCESS_ATTACH:

            // If the DLL matches our list of modules, apply the hotpatch
            
            struct module* module = module_match(hinstDll);
            if (module != NULL)
                patch_module(module, hinstDll);
            break;
    }

    return;
}


/* To intercept DLL loading and unloading, we will hook the ntdll function that
   calls the DllMain functions for all DLLs. The body of this function is
   exactly the same on Windows 2000, XP and 2003.

.text:7C901193             _LdrpCallInitRoutine@16 proc near
.text:7C901193
.text:7C901193             arg_0           = dword ptr  8
.text:7C901193             arg_4           = dword ptr  0Ch
.text:7C901193             arg_8           = dword ptr  10h
.text:7C901193             arg_C           = dword ptr  14h
.text:7C901193
.text:7C901193 55                          push    ebp
.text:7C901194 8B EC                       mov     ebp, esp
.text:7C901196 56                          push    esi
.text:7C901197 57                          push    edi
.text:7C901198 53                          push    ebx
.text:7C901199 8B F4                       mov     esi, esp
.text:7C90119B FF 75 14                    push    [ebp+arg_C]
.text:7C90119E FF 75 10                    push    [ebp+arg_8]
.text:7C9011A1 FF 75 0C                    push    [ebp+arg_4]
.text:7C9011A4 FF 55 08                    call    [ebp+arg_0]
.text:7C9011A7 8B E6                       mov     esp, esi
.text:7C9011A9 5B                          pop     ebx
.text:7C9011AA 5F                          pop     edi
.text:7C9011AB 5E                          pop     esi
.text:7C9011AC 5D                          pop     ebp
.text:7C9011AD C2 10 00                    retn    10h
.text:7C9011AD             _LdrpCallInitRoutine@16 endp

   This function is not exported, but there is a simple way to get its address.
   When our own DllMain function is called, the return address of the stack
   will point to the mov esp, esi instruction in LdrpCallInitRoutine. We will
   verify that the 12 bytes before the return address match, and then we will
   overwrite the first two push instructions with a jump to our loader hook
   function. The loader hook function will push the three DllMain arguments,
   call DllMainHook, push the three arguments again and jump back to the
   call [ebp+arg_0] instrcution in ntdll.dll. */


// The 12 bytes at the patch point are three push instructions and a call

const char loader_patch_bytes[] = "\xFF\x75\x14"    // push [ebp+arg_C]
                                  "\xFF\x75\x10"    // push [ebp+arg_8]
                                  "\xFF\x75\x0C"    // push [ebp+arg_4]
                                  "\xFF\x55\x08";   // call [ebp+arg_0]

const int  loader_patch_size = sizeof(loader_patch_bytes)-1;

// Patch point address, NULL if the loader hook is not currently inserted

char* loader_patch_point = NULL;

// This is the address of the call [ebp+arg_0] instrcution in ntdll.dll. The
// loader hook function reads this variable to find out where to return to.

char* loader_hook_return;


//
// Loader hook function, called from LdrpCallInitRoutine
//

__declspec(naked) void loader_hook()
{
    __asm {
        push [ebp+0x14]             // push [ebp+arg_C]
        push [ebp+0x10]             // push [ebp+arg_8]
        push [ebp+0xC]              // push [ebp+arg_4]
        call DllMainHook            // call our DllMain hook
        
        push [ebp+0x14]             // push [ebp+arg_C]
        push [ebp+0x10]             // push [ebp+arg_8]
        push [ebp+0xC]              // push [ebp+arg_4]
        jmp  [loader_hook_return]   // jump to the call [ebp+arg_0] instruction
    }
}

// Relative jump opcode

#define JMP_REL32 '\xE9'

//
// Patch the ntdll!LdrpCallInitRoutine function. The argument is the return
// address from DllMain. Returns FALSE if the patch could not be applied.
//

BOOL insert_loader_hook(char* return_address)
{
    char* patch_point = return_address - loader_patch_size;

    if (memcmp(patch_point, loader_patch_bytes, loader_patch_size) != 0) {
        return FALSE;
    }

    // Store the address of the call [ebp+arg_0] instruction

    loader_hook_return = return_address - 3;

    // Change the memory protection of the patch point to enable writing
    
    DWORD old_protection, dummy;

    if (VirtualProtect(patch_point, 5, PAGE_EXECUTE_READWRITE, &old_protection) == 0) {
        return FALSE;
    }

    // Write a 5 byte relative jump to the loader hook

    int loader_hook_rel = (int)&loader_hook - ((int)patch_point + 5);

            patch_point[0] = JMP_REL32;         // relative jump opcode
    *(int*)&patch_point[1] = loader_hook_rel;   // relative jump target

    loader_patch_point = patch_point;
    
    // Restore the memory protection of the patch point

    VirtualProtect(patch_point, 5, old_protection, &dummy);

    return TRUE;
}


//
// Remove the loader hook, returns FALSE on error
//

BOOL remove_loader_hook()
{
    if (loader_patch_point != NULL) {

        // Change the memory protection of the patch point to enable writing
        
        DWORD old_protection, dummy;

        if (VirtualProtect(loader_patch_point, 5, PAGE_EXECUTE_READWRITE, &old_protection) == 0) {
            return FALSE;
        }

        // Restore the 5 bytes that we overwrote

        memcpy(loader_patch_point, loader_patch_bytes, 5);

        // Restore the memory protection of the patch point

        VirtualProtect(loader_patch_point, 5, old_protection, &dummy);
    }

    return TRUE;
}


//
// DLL entry point
//

BOOL WINAPI DllMain(HINSTANCE hinstDll, DWORD fdwReason, LPVOID lpvReserved)
{
    UNREFERENCED_PARAMETER(hinstDll);
    UNREFERENCED_PARAMETER(lpvReserved);

    // Get the return address of the current function
    char* return_address = ((char**)&hinstDll)[-1];

    switch (fdwReason) {
        case DLL_PROCESS_ATTACH:

            // patch the DLLs that have already been loaded
            patch_loaded_modules();

            // insert the loader hook
            insert_loader_hook(return_address);
            break;

        case DLL_PROCESS_DETACH:
            // remove the loader hook
            remove_loader_hook();
            break;
    }

    return TRUE;
}
